import pathlib

from blissoda.bliss_globals import setup_globals
from blissoda.demo import testing
from blissoda.demo.processors.id14 import DemoId14Hdf5ToSpecConverter

id14_converter = DemoId14Hdf5ToSpecConverter()


def id14_demo(expo=0.2, npoints=10):
    for _ in range(2):
        test_id14_loopscan(expo=expo, npoints=npoints)


@testing.integration_fixture
def _id14():
    id14_converter.enable()
    yield id14_converter
    id14_converter.disable()


@testing.integration_test
def test_id14_loopscan(_id14, expo=0.2, npoints=10):
    scan = setup_globals.loopscan(
        npoints, expo, setup_globals.diode1, setup_globals.mca1
    )
    scan_number = scan.scan_info["scan_nb"]
    _assert_spec_counter_file(_id14, scan_number)
    _assert_spec_mca_file(_id14, scan_number)


@testing.demo_assert("Check ID14 SPEC file for scan #{scan_number}")
def _assert_spec_counter_file(_id14, scan_number):
    result = _id14._future_for_counters.result(timeout=10)
    output_filename = pathlib.Path(result["output_filename"])
    testing.assert_spec_scan_exists(output_filename, scan_number)


@testing.demo_assert("Check ID14 MCA file for scan #{scan_number}")
def _assert_spec_mca_file(_id14, scan_number):
    result = _id14._future_for_mca.result(timeout=10)
    output_filenames = result["output_filenames"]
    if not output_filenames:
        raise AssertionError("No filenames returned from ID14 MCA conversion workflow")
    for output_filename in output_filenames:
        testing.assert_spec_scan_exists(output_filename, scan_number)



from __future__ import annotations

from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Sequence

import numpy as np

from ...flint import WithFlintAccess
from ...flint import capture_errors


class GridPlot(WithFlintAccess):
    """Keep a named Flint grid tab alive and easy to update"""

    def __init__(
        self, unique_name: str = "grid-plot", window_title: str = "Energy Calculation"
    ) -> None:
        super().__init__()
        self._unique_name = unique_name
        self._window_title = window_title
        self._rows = 0
        self._cols = 0

    def _container(self, *, select: bool = True):
        return self._flint_client.get_plot(
            "grid",
            name=self._window_title,
            unique_name=self._unique_name,
            selected=select,
        )

    @staticmethod
    def _as_float_array(a) -> np.ndarray:
        return (
            np.asarray(a, dtype=float) if a is not None else np.array([], dtype=float)
        )

    @staticmethod
    def _clean_plot(p) -> None:
        clean = getattr(p, "clean_data", None)
        if callable(clean):
            clean()

    @staticmethod
    def _set_plot_labels(
        p, *, title: Optional[str], xlabel: Optional[str], ylabel: Optional[str]
    ) -> None:
        if title is not None:
            p.title = title
        if xlabel is not None:
            p.xlabel = xlabel
        if ylabel is not None:
            p.ylabel = ylabel

    @staticmethod
    def _add_curve(
        p,
        x: np.ndarray,
        y: np.ndarray,
        *,
        legend: str,
        color: Optional[str] = None,
        linewidth: Optional[float] = None,
    ) -> None:
        kwargs = {"legend": legend}
        if color is not None:
            kwargs["color"] = color
        if linewidth is not None:
            kwargs["linewidth"] = linewidth
        p.add_curve(x, y, **kwargs)

    @capture_errors
    def clear(self) -> None:
        """Clear all known cells."""
        if not (self._rows and self._cols):
            return
        cont = self._container(select=False)
        for r in range(self._rows):
            for c in range(self._cols):
                p = cont.get_plot("curve", row=r, col=c)
                self._clean_plot(p)

    @capture_errors
    def set_layout(self, rows: int, cols: int) -> None:
        """Remember target grid size and pre-create visible cells."""
        rows, cols = max(1, int(rows)), max(1, int(cols))
        self._rows, self._cols = rows, cols
        cont = self._container(select=True)
        for r in range(rows):
            for c in range(cols):
                p = cont.get_plot("curve", row=r, col=c)
                self._clean_plot(p)

    @capture_errors
    def set_cell(
        self,
        row: int,
        col: int,
        *,
        x,
        series: Sequence[Mapping],
        title: Optional[str] = None,
        xlabel: Optional[str] = None,
        ylabel: Optional[str] = None,
    ) -> None:
        """Update a single cell (row, col) with one or more curves."""
        r, c = int(row), int(col)
        cont = self._container(select=True)
        plot = cont.get_plot("curve", row=r, col=c)
        self._clean_plot(plot)

        self._set_plot_labels(plot, title=title, xlabel=xlabel, ylabel=ylabel)

        x_arr = None if x is None else self._as_float_array(x)
        for s in series or []:
            y = s.get("y")
            if y is None:
                continue
            y_arr = self._as_float_array(y)

            if x_arr is None:
                xx = np.arange(y_arr.size, dtype=float)
            else:
                n = min(x_arr.size, y_arr.size)
                if n == 0:
                    continue
                xx = x_arr[:n]
                y_arr = y_arr[:n]

            self._add_curve(
                plot,
                xx,
                y_arr,
                legend=s.get("label", "series"),
                color=s.get("color"),
                linewidth=s.get("linewidth"),
            )

        self._rows = max(self._rows, r + 1)
        self._cols = max(self._cols, c + 1)

    @capture_errors
    def set_cells(self, batch: Iterable[Mapping]) -> None:
        """Batch update: iterable of dicts accepted by set_cell."""
        for item in batch:
            d = dict(item)
            self.set_cell(
                int(d.get("row", 0)),
                int(d.get("col", 0)),
                x=d.get("x"),
                series=d.get("series") or (),
                title=d.get("title"),
                xlabel=d.get("xlabel"),
                ylabel=d.get("ylabel"),
            )



import logging
from functools import wraps
from importlib.metadata import version
from typing import Callable
from typing import Type

from packaging.version import Version

from ..bliss_globals import current_session
from ..import_utils import is_available
from ..import_utils import unavailable_class
from ..import_utils import unavailable_function

try:
    from gevent.lock import RLock
except ImportError as ex:
    RLock = unavailable_class(ex)

try:
    from bliss.common.plot import get_flint
except ImportError as ex:
    get_flint = unavailable_function(ex)

try:
    if Version(version("bliss")) >= Version("2.2"):
        from flint.client.plots import BasePlot
        from flint.client.proxy import FlintClient
    else:
        from bliss.flint.client.plots import BasePlot
        from bliss.flint.client.proxy import FlintClient
except ImportError as ex:
    FlintClient = unavailable_class(ex)
    BasePlot = unavailable_class(ex)


logger = logging.getLogger(__name__)


class WithFlintAccess:
    _HAS_BLISS = is_available(current_session)

    def __init__(self) -> None:
        self._client = None
        self.__clientlock = None
        self._plots = dict()

    @property
    def _clientlock(self) -> RLock:
        if self.__clientlock is None:
            self.__clientlock = RLock()
        return self.__clientlock

    def _get_plot(self, plot_id: str, plot_cls: Type[BasePlot]) -> BasePlot:
        """Launches Flint and creates the plot when either is missing"""
        plot = self._plots.get(plot_id)
        if plot is None:
            plot = self._flint_client.get_plot(plot_cls, unique_name=plot_id)
            self._plots["plot_id"] = plot
        return plot

    @property
    def _flint_client(self) -> FlintClient:
        """Launches Flint when missing"""
        with self._clientlock:
            try:
                if self._client.is_available():
                    return self._client
            except (FileNotFoundError, AttributeError):
                pass
            self._client = get_flint()
            self._plots = dict()
            return self._client


def capture_errors(method) -> Callable:
    @wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            msg = f"Flint plot error: {e}"
            logger.error(msg, exc_info=True)

    return wrapper



import time
from typing import Any

from ewoksjob.client import Future
from ewoksjob.client.celery.utils import get_not_finished_futures

from ._display import print_message


def wait_workflows(*futures: Future, timeout: int = 30) -> Any:
    if not futures:
        futures = get_not_finished_futures()

    t0 = time.time()
    for future in futures:
        timepassed = time.time() - t0
        _ = future.result(timeout=timeout - timepassed)
        print_message(f"Job {future.uuid!r} finished", "info")
